// Info.cpp

#include "stdafx.h"
#include <ddraw.h>
#include "dd.h"
#include "resource.h"

#define PIXEL_FLAG_MASK	(DDPF_PALETTEINDEXED8 | DDPF_RGB)

static DWORD	g_dwCharWidth;
static DWORD	g_dwCharHeight;
static TCHAR	g_szInfoString[256];
static DWORD	g_dwCount = UINT_MAX;
static DWORD	g_dwPixelFlag = 0xFFFFFFFF;
static DWORD	g_dwBitCount = 0;
static DWORD	g_dwWidth;
static DWORD	g_dwHeight;
static int		g_fps;
static DWORD	g_dwFpsCount = UINT_MAX;
static BITMAPINFOHEADER		g_bmih;
static LPDIRECTDRAWSURFACE	lpBmpSurface = NULL;

extern BOOL	bEraseBackground;

static void DispChar(LPDIRECTDRAWSURFACE lpDst, int xPos, int yPos, LPDIRECTDRAWSURFACE lpSrc, int c);
static BOOL LoadBmp(DDSURFACEDESC *pddsd);
static void BmpToDDSurface(LPDIRECTDRAWSURFACE lpSurface, void *pBmp);
static void GetBmpInfo(BITMAPINFOHEADER *pbmih, void *pBmp);

void InitInfo(void)
{
}

void TrashInfo(void)
{
	if (lpBmpSurface != NULL)
		lpBmpSurface->Release();
}

void SetInfoString(LPCTSTR szInfo)
{
	strcpy(g_szInfoString, szInfo);
	g_dwCount = 0;
}

void SetInfoFps(int fps)
{
	if (fps != g_fps)
	{
		g_fps = fps;
		bEraseBackground = TRUE;
	}
	g_dwFpsCount = 0;
}

#define INFO_TIME	120

void DisplayInfo(LPDIRECTDRAWSURFACE lpSurface)
{
	HRESULT	ddret;
	int		xPos, yPos;
	int		c;
	int		i;
	int		len;
	DDSURFACEDESC	ddsd;

//	if (g_szInfoString[0] == _T('\0'))
//		return;
	if (g_dwCount == INFO_TIME)
	{
		g_dwCount++;
		bEraseBackground = TRUE;
	}
	if (g_dwFpsCount == 1)
	{
		g_dwFpsCount++;
		bEraseBackground = TRUE;
	}
	if (g_dwCount >= INFO_TIME && g_dwFpsCount >= 1)
		return;

	memset(&ddsd, 0, sizeof(DDSURFACEDESC));
	ddsd.dwSize = sizeof(ddsd);
	ddret = lpSurface->Lock(NULL, &ddsd, 0, NULL);
	if (FAILED(ddret))
	{
		TRACE0("Lock failed\n");
		return;
	}
	lpSurface->Unlock(NULL);
	if (g_dwPixelFlag != (ddsd.ddpfPixelFormat.dwFlags & PIXEL_FLAG_MASK) ||
		(ddsd.ddpfPixelFormat.dwFlags & DDPF_RGB) && g_dwBitCount != ddsd.ddpfPixelFormat.dwRGBBitCount ||
		(ddsd.dwHeight != g_dwHeight || ddsd.dwWidth != g_dwWidth))
	{
		if (!LoadBmp(&ddsd))
			return;
	}
	else if (lpBmpSurface->IsLost() == DDERR_SURFACELOST)
	{
		ddret = lpBmpSurface->Restore();
		if (FAILED(ddret))
		{
			TRACE0("Restore failed\n");
			return;
		}
		if (!LoadBmp(&ddsd))
			return;
	}

	if (g_dwCount < INFO_TIME)
	{
		g_dwCount++;
		xPos = 8;
		yPos = ddsd.dwHeight - (g_dwCharHeight+4)*2 - 4;
		len = _tcslen(g_szInfoString);
		for (i = 0; i < len; i++)
		{
			c = g_szInfoString[i];
			if (c < 32)
			{
				// newline upward
				if (c == _T('\n'))
					yPos -= g_dwCharHeight + 4;
			}
			else if (c < 128)
			{
				DispChar(lpSurface, xPos, yPos, lpBmpSurface, c);
				xPos += g_dwCharWidth;
			}
		}
	}

	if (g_dwFpsCount < 1)
	{
		TCHAR	szBuf[16];

		g_dwFpsCount++;
		xPos = 8;
		yPos = 8;
		wsprintf(szBuf, _T("%3d"), g_fps);
		len = _tcslen(szBuf);
		for (i = 0; i < len; i++)
		{
			DispChar(lpSurface, xPos, yPos, lpBmpSurface, szBuf[i]);
			xPos += g_dwCharWidth;
		}
	}
}

void DispChar(LPDIRECTDRAWSURFACE lpDst, int xPos, int yPos, LPDIRECTDRAWSURFACE lpSrc, int c)
{
	RECT	rcDst, rcSrc;

	rcDst.left = xPos;
	rcDst.top = yPos;
	rcDst.right = xPos + g_dwCharWidth;
	rcDst.bottom = yPos + g_dwCharHeight;
	rcSrc.left = ((c-32) & 31)*g_dwCharWidth;
	rcSrc.top = ((c-32) >> 5)*g_dwCharHeight;
	rcSrc.right = rcSrc.left + g_dwCharWidth;
	rcSrc.bottom = rcSrc.top + g_dwCharHeight;
	lpDst->BltFast(xPos, yPos, lpSrc, &rcSrc, DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT);
}

BOOL LoadBmp(DDSURFACEDESC *pddsd)
{
	HRSRC	hRsrc;
	HGLOBAL	hResData;
	void	*pBmp;
	DDSURFACEDESC	ddsdChar;
	HRESULT	ddret;

	g_dwPixelFlag = pddsd->ddpfPixelFormat.dwFlags & PIXEL_FLAG_MASK;
	g_dwBitCount = pddsd->ddpfPixelFormat.dwRGBBitCount;
	g_dwWidth = pddsd->dwWidth;
	g_dwHeight = pddsd->dwHeight;
	hRsrc = FindResource(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDR_CHARS_BMP), _T("BITMAPFILE"));
	if (hRsrc == NULL)
	{
		TRACE("FindResource failed\n");
		return FALSE;
	}
	hResData = LoadResource(AfxGetInstanceHandle(), hRsrc);
	if (hResData == NULL)
	{
		TRACE("LoadResource failed\n");
		return FALSE;
	}
	pBmp = LockResource(hResData);
	if (pBmp == NULL)
	{
		TRACE("LoadResource failed\n");
		return FALSE;
	}
	GetBmpInfo(&g_bmih, pBmp);

	if (lpBmpSurface != NULL)
		lpBmpSurface->Release();

	ddsdChar.dwSize = sizeof(ddsdChar);
	ddsdChar.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsdChar.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsdChar.dwWidth = g_bmih.biWidth;
	ddsdChar.dwHeight = g_bmih.biHeight;
	ddret = lpDDraw->CreateSurface(&ddsdChar, &lpBmpSurface, NULL);
	if (FAILED(ddret))
	{
		TRACE0("CreateSurfaces\n");
		return FALSE;
	}

	BmpToDDSurface(lpBmpSurface, pBmp);
	g_dwCharWidth = g_bmih.biWidth/32;
	g_dwCharHeight = g_bmih.biHeight/3;

	return TRUE;
}

void GetBmpInfo(BITMAPINFOHEADER *pbmih, void *pBmp)
{
	*pbmih = *(BITMAPINFOHEADER *)(((BITMAPFILEHEADER *)pBmp)+1);
}

static DWORD GetPattern(DDPIXELFORMAT *pddpf);
static DWORD GetColorKey(DDPIXELFORMAT *pddpf);

void BmpToDDSurface(LPDIRECTDRAWSURFACE lpSurface, void *pBmp)
{
	BITMAPFILEHEADER	*pbmfh;
	BITMAPINFOHEADER	*pbmih;
	DDSURFACEDESC	ddsd;
	HRESULT		ddret;
	BYTE		*pBits;
	BYTE		*lpDDMemory;
	BYTE		*pd;
	BYTE		*ps;
	int			nBytes;
	DWORD		dwData;
	DWORD		dwColorKey;
	DWORD		w, h;
	DWORD		c;
	int			j;
	int			iPitch;

	pbmfh = (BITMAPFILEHEADER *)pBmp;
	pbmih = (BITMAPINFOHEADER *)(pbmfh + 1);
	if (pbmih->biBitCount != 8)
	{
		TRACE("Bitmap is not 256color\n");
		return;
	}
	pBits = (BYTE *)pBmp + pbmfh->bfOffBits;

	memset(&ddsd, 0, sizeof(DDSURFACEDESC));
	ddsd.dwSize = sizeof(ddsd);
	ddret = lpSurface->Lock(NULL, &ddsd, 0, NULL);
	if (FAILED(ddret))
	{
		TRACE0("Locks\n");
		return;
	}

	if (ddsd.ddpfPixelFormat.dwFlags & DDPF_PALETTEINDEXED8)
		nBytes = 1;
	else if (ddsd.ddpfPixelFormat.dwFlags & DDPF_RGB)
		nBytes = ddsd.ddpfPixelFormat.dwRGBBitCount/8;
	dwData = GetPattern(&ddsd.ddpfPixelFormat);
	dwColorKey = GetColorKey(&ddsd.ddpfPixelFormat);

	iPitch = ((pbmih->biWidth-1) & ~3)+4;
	pBits += iPitch * (pbmih->biHeight-1);
	lpDDMemory = (BYTE *)ddsd.lpSurface;

	for (h = 0; h < ddsd.dwHeight; h++)
	{
		pd = lpDDMemory;
		ps = pBits;
		for (w = 0; w < ddsd.dwWidth; w++)
		{
			c = *ps++;
			if (c == 1)
			{
				for (j = 0; j < nBytes; j++)
					*pd++ = ((BYTE *)&dwData)[j];
			}
			else if (c == 0)
			{
				for (j = 0; j < nBytes; j++)
					*pd++ = 0;
			}
			else
			{
				for (j = 0; j < nBytes; j++)
					*pd++ = ((BYTE *)&dwColorKey)[j];
			}
		}
		lpDDMemory += ddsd.lPitch;
		pBits -= iPitch;
	}
	lpSurface->Unlock(NULL);
	TRACE("dwData = %X\n", dwData);
	TRACE("dwColorKey = %X\n", dwColorKey);

	DDCOLORKEY	ddck;
	ddck.dwColorSpaceLowValue = dwColorKey;
	ddck.dwColorSpaceHighValue = dwColorKey;
	lpSurface->SetColorKey(DDCKEY_SRCBLT, &ddck);
}

DWORD GetPattern(DDPIXELFORMAT *pddpf)
{
	if (pddpf->dwFlags & DDPF_PALETTEINDEXED8)
		return 255;
	else if (pddpf->dwFlags & DDPF_RGB)
		return 0xFFFFFFFF & (pddpf->dwRBitMask | pddpf->dwGBitMask | pddpf->dwBBitMask);
	else
		return 0xFFFFFFFF;
}

DWORD GetColorKey(DDPIXELFORMAT *pddpf)
{
	if (pddpf->dwFlags & DDPF_PALETTEINDEXED8)
		return 2;
	else if (pddpf->dwFlags & DDPF_RGB)
	{
		DWORD r = (pddpf->dwRBitMask ^ (pddpf->dwRBitMask>>1)) & pddpf->dwRBitMask;
		DWORD g = (pddpf->dwGBitMask ^ (pddpf->dwGBitMask>>1)) & pddpf->dwGBitMask;
		DWORD b = (pddpf->dwBBitMask ^ (pddpf->dwBBitMask>>1)) & pddpf->dwBBitMask;
		return r|g|b;
	}
	else
		return 2;
}
