// dd.cpp

#include "stdafx.h"
#include <ddraw.h>
#include "MainFrame.h"
#include "info.h"
#include "log.h"
#define DD_CPP
#include "dd.h"

LPDIRECTDRAW	lpDDrawOld = NULL;
LPDIRECTDRAW2	lpDDraw = NULL;
LPDIRECTDRAWSURFACE	lpPrimarySurface = NULL;
LPDIRECTDRAWSURFACE	lpBackSurface = NULL;
LPDIRECTDRAWCLIPPER	lpClipper = NULL;
//LPDIRECTDRAWSURFACE	lpSurface;

SCREEN_MODE	mode[MAX_MODE] =
{
	{320, 240, 1},
	{640, 480, 2},
};
int curMode = 0;
int numMode = 2;
int xPosWindowed = 0;
int yPosWindowed = 0;

BOOL bFullScreen = FALSE;
BOOL bActiveApp;
//int zoom = 1;

HPALETTE hPaletteApp;

HWND hwnd = NULL;
HMENU hMenu = NULL;

RECT rcWnd;

BOOL InitDDraw(void)
{
	HRESULT	ddret;

	LogDumpDetail("DD:DirectDrawCreate...");
	ddret = DirectDrawCreate(NULL, &lpDDrawOld, NULL);
	LogDumpDetail("Done\n");
	if (FAILED(ddret))
	{
		LogDump("InitDDraw - DirectDrawCreates [code=%08X]\n", ddret);
		return FALSE;
	}
	LogDumpDetail("DD:QueryInterface...");
	ddret = lpDDrawOld->QueryInterface(IID_IDirectDraw2, (LPVOID *)&lpDDraw);
	LogDumpDetail("Done\n");
	if(FAILED(ddret))
	{
		LogDump("InitDDraw - IDirectDraw::QueryInterfaces [code=%08X]\n", ddret);
	    return FALSE;
	}
	lpDDrawOld->Release();

	return TRUE;
}

void SetPalette(void)
{
	static struct pal
	{
		WORD	palVersion;
		WORD	palNumEntries;
		PALETTEENTRY	pe[256];
	} logpal;
	PALETTEENTRY *const &pe = (PALETTEENTRY *)logpal.pe;
	LPDIRECTDRAWPALETTE	lpDDPalette;
	int		i;
	HRESULT	ddret;

	if (bFullScreen)
	{
		for(i = 0; i < 256; i++)
		{
			pe[i].peRed = ((i>>2)&7)*36;
			pe[i].peGreen = (i>>5)*36;
			pe[i].peBlue = (i&3)*84;
		}

		LogDumpDetail("DD:CreatePalette...");
		ddret = lpDDraw->CreatePalette(DDPCAPS_8BIT, pe, &lpDDPalette, NULL);
		LogDumpDetail("Done\n");
		if (FAILED(ddret))
		{
			LogDump("SetPalette - CreatePalettes [code=%08X]\n", ddret);
			return;
		}

		LogDumpDetail("DD:SetPalette...");
		ddret = lpPrimarySurface->SetPalette(lpDDPalette);
		LogDumpDetail("Done\n");
		if (FAILED(ddret))
		{
			LogDump("SetPalette - SetPalettes [code=%08X]\n", ddret);
			return;
		}
		lpDDPalette->Release();
	}
	else
	{
		for(i = 0; i < 256; i++)
		{
			pe[i].peRed = 0;
			pe[i].peGreen = 0;
			pe[i].peBlue = 0;
			pe[i].peFlags = 0;//PC_NOCOLLAPSE;
		}
		HDC hdc = GetDC(NULL);
		HPALETTE hPalette = CreatePalette((LOGPALETTE *)&logpal);
		hPalette = SelectPalette(hdc, hPalette, FALSE);
		RealizePalette(hdc);
		hPalette = SelectPalette(hdc, hPalette, FALSE);
		DeleteObject(hPalette);

		GetSystemPaletteEntries(hdc, 0, 256, pe);
		ReleaseDC(NULL, hdc);

/*		for (i = 0; i < 256; i++)
		{
			pe[i].peRed = 0;
			pe[i].peGreen = 0;
			pe[i].peBlue = 0;
		}
*/
		for(i = 10; i < 246; i++)
		{
			pe[i].peRed = ((i>>2)&7)*36;
			pe[i].peGreen = (i>>5)*36;
			pe[i].peBlue = (i&3)*84;
		}

		logpal.palVersion = 0x300;
		logpal.palNumEntries = 256;
		for (i = 0; i < 256; i++)
			pe[i].peFlags = 0;//PC_NOCOLLAPSE;
		if (hPalette)
			DeleteObject(hPalette);
		hPaletteApp = CreatePalette((LOGPALETTE *)&logpal);

		PostMessage(hwnd, WM_QUERYNEWPALETTE, 0, 0);
	}
}

BOOL InitSurfaces(BOOL bSetFullScreen)
{
	HRESULT	ddret;
	DDSURFACEDESC	ddsd;

	TRACE("SetCooperativeLevel: fullscreen = %d\n", bSetFullScreen);

	lpDDraw->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, NULL);
	LogDumpDetail("DD:SetCooperativeLevel...");
	if (bSetFullScreen)
		ddret = lpDDraw->SetCooperativeLevel(hwnd, DDSCL_ALLOWREBOOT | DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
	else
		ddret = lpDDraw->SetCooperativeLevel(hwnd, DDSCL_NORMAL);
	LogDumpDetail("Done\n");
	if (FAILED(ddret))
	{
		LogDump("InitSurfaces - SetCooperativeLevels [code=%08X]\n", ddret);
		return FALSE;
	}

	bFullScreen = bSetFullScreen;

	TRACE("SetDisplayMode\n");

	if (bFullScreen)
	{
		TRACE("w = %d, h = %d\n", mode[curMode].width, mode[curMode].height);
		lpDDraw->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, NULL);
		LogDumpDetail("DD:SetDisplayMode...");
		ddret = lpDDraw->SetDisplayMode(mode[curMode].width, mode[curMode].height, 8, 0, 0);
//		ddret = lpDDraw->SetDisplayMode(512, 384, 8, 0, 0);
		LogDumpDetail("Done\n");
		if (FAILED(ddret))
		{
			LogDump("InitSurfaces - SetDisplayModes [code=%08X]\n", ddret);
			lpDDraw->SetCooperativeLevel(hwnd, DDSCL_NORMAL);
			bFullScreen = FALSE;
		}
	}

	TRACE("CreateSurface\n");

	ddsd.dwSize = sizeof(ddsd);
	if (bFullScreen)
	{
		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
		ddsd.dwBackBufferCount = 1;
	}
	else
	{
		ddsd.dwFlags = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	}
	LogDumpDetail("DD:CreateSurface...");
	ddret = lpDDraw->CreateSurface(&ddsd, &lpPrimarySurface, NULL);
	LogDumpDetail("Done\n");
	if (FAILED(ddret))
	{
		LogDump("InitSurfaces - CreateSurfaces [code=%08X]\n", ddret);
		return FALSE;
	}

	if (!bFullScreen)
	{
		LogDumpDetail("DD:CreateClipper...");
		ddret = lpDDraw->CreateClipper(0, &lpClipper, NULL);
		LogDumpDetail("Done\n");
		if (FAILED(ddret))
		{
			LogDump("InitSurfaces - CreateClippers [code=%08X]\n", ddret);
			return FALSE;
		}
		LogDumpDetail("DD:SetHWnd...");
		ddret = lpClipper->SetHWnd(0, hwnd);
		LogDumpDetail("Done\n");
		if (FAILED(ddret))
		{
			LogDump("InitSurfaces - SetHWnds [code=%08X]\n", ddret);
			return FALSE;
		}
		LogDumpDetail("DD:SetClipper...");
		ddret = lpPrimarySurface->SetClipper(lpClipper);
		LogDumpDetail("Done\n");
		if (FAILED(ddret))
		{
			LogDump("InitSurfaces - SetClippers [code=%08X]\n", ddret);
			return FALSE;
		}
	}

	if (bFullScreen)
	{
		DDSCAPS	ddscaps;
		ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
		LogDumpDetail("DD:GetAttachedSurface...");
		ddret = lpPrimarySurface->GetAttachedSurface(&ddscaps, &lpBackSurface);
		LogDumpDetail("Done\n");
		if (FAILED(ddret))
		{
			LogDump("InitSurfaces - GetAttachedSurfaces [code=%08X]\n", ddret);
			return FALSE;
		}
	}
	else
	{
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
		ddsd.dwWidth = 320*((curMode&1)+1);
		ddsd.dwHeight = 240*((curMode&1)+1);
		LogDumpDetail("DD:CreateSurface(back)...");
		ddret = lpDDraw->CreateSurface(&ddsd, &lpBackSurface, NULL);
		LogDumpDetail("Done\n");
		if (FAILED(ddret))
		{
			LogDump("InitSurfaces - CreateSurface(back)s [code=%08X]\n", ddret);
			return FALSE;
		}
	}

	SetPalette();

#if 0
	if (!bFullScreen)
	{
		TRACE("Creating surface for VRAM dumping.\n");
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
		ddsd.dwHeight = 640;
		ddsd.dwWidth = 640;
		ddret = lpDDraw->CreateSurface(&ddsd, &lpSurface, NULL);
		if (FAILED(ddret))
		{
			TRACE("CreateSurface Failed\n");
			return FALSE;
		}
	}
#endif

	return TRUE;
}

BOOL RestoreSurfaces(void)
{
	HRESULT	ddret;

	ddret = lpPrimarySurface->Restore();
	if (FAILED(ddret))
	{
		LogDumpDetail("Restore(p) failed,code=%08X\n",ddret);
		return FALSE;
	}

	ddret = lpBackSurface->Restore();
	if (FAILED(ddret))
	{
		LogDumpDetail("Restore(b) failed,code=%08X\n",ddret);
		return FALSE;
	}

	return TRUE;
}

BOOL ReleaseSurfaces(void)
{
	if (lpDDraw != NULL)
	{
		if (lpClipper)
		{
			LogDumpDetail("DD:Release(clipper)...");
			lpClipper->Release();
			LogDumpDetail("Done\n");
			lpClipper = NULL;
		}
		if (lpBackSurface)
		{
			LogDumpDetail("DD:Release(backsurface)...");
			lpBackSurface->Release();
			LogDumpDetail("Done\n");
			lpBackSurface = NULL;
		}
		if (lpPrimarySurface)
		{
			LogDumpDetail("DD:Release(primarysurface)...");
			lpPrimarySurface->Release();
			LogDumpDetail("Done\n");
			lpPrimarySurface = NULL;
		}
//		if (lpSurface)
//		{
//			lpSurface->Release();
//			lpSurface = NULL;
//		}
	}

	return TRUE;
}

BOOL ReleaseDDraw(void)
{
	if (lpDDraw)
	{
		LogDumpDetail("DD:SetCooperativeLevel...");
		lpDDraw->SetCooperativeLevel(hwnd, DDSCL_NORMAL);
		LogDumpDetail("Done\nDD:Release(ddraw)...");
		lpDDraw->Release();
		LogDumpDetail("Done\n");
		lpDDraw = NULL;
	}
	if (hPaletteApp)
	{
		DeleteObject(hPaletteApp);
		hPaletteApp = NULL;
	}
	return TRUE;
}

void ChangeWindowMode(void)
{
	TCHAR	szBuf[16];

	bFullScreen = !bFullScreen;

	if (bFullScreen)
	{
		RECT	rc;
		GetWindowRect(hwnd, &rc);
		xPosWindowed = rc.left;
		yPosWindowed = rc.top;

		ShowCursor(FALSE);
		hMenu = GetMenu(hwnd);
		SetMenu(hwnd, NULL);
		SetWindowLong(hwnd, GWL_STYLE, WS_POPUP);
		SetWindowLong(hwnd, GWL_EXSTYLE, WS_EX_TOPMOST);
		MoveWindow(hwnd, 0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN), TRUE);
		ReleaseSurfaces();
//		lpDDraw->Release();
//		AfxMessageBox("Test");
//		if (!InitDDraw())
//			TRACE("Cannot InitDDraw\n");
		InitSurfaces(bFullScreen);
		wsprintf(szBuf, _T("%dx%dx8"), mode[curMode].width, mode[curMode].height);
		SetInfoString(szBuf);
	}
	if (!bFullScreen)
	{
		ReleaseSurfaces();
//		lpDDraw->Release();
//		if (!InitDDraw())
//			TRACE("Cannot InitDDraw\n");
		//curMode &= 1;
		InitSurfaces(bFullScreen);
		SetWindowLong(hwnd, GWL_STYLE, WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_BORDER | WS_MINIMIZEBOX);
		SetWindowLong(hwnd, GWL_EXSTYLE, WS_EX_CLIENTEDGE);
		SetWindowPos(hwnd, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE | SWP_SHOWWINDOW);
		int cx = 320*((curMode&1)+1) + GetSystemMetrics(SM_CXFIXEDFRAME)*2 + GetSystemMetrics(SM_CXEDGE)*2;
		int cy = 240*((curMode&1)+1) + GetSystemMetrics(SM_CYFIXEDFRAME)*2 + GetSystemMetrics(SM_CYEDGE)*2
					+ GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU);
		MoveWindow(hwnd, xPosWindowed, yPosWindowed, cx, cy, TRUE);
		ShowCursor(TRUE);
		SetMenu(hwnd, hMenu);
		wsprintf(szBuf, _T("%dx%d"), 320*((curMode&1)+1), 240*((curMode&1)+1));
		SetInfoString(szBuf);
	}

	HRESULT	ddret;
	DDSURFACEDESC	ddsd;
	BYTE	*lpDDMemory;

	memset(&ddsd, 0, sizeof(DDSURFACEDESC));
	ddsd.dwSize = sizeof(ddsd);
	LogDumpDetail("DD:Lock(winmode)...");
	ddret = lpBackSurface->Lock(NULL, &ddsd, 0, NULL);
	LogDumpDetail("Done\n");
	if (FAILED(ddret))
	{
		LogDump("ChangeWindowMode - Locks [code=%08X]\n", ddret);
		return;
	}

	lpDDMemory = (BYTE *)ddsd.lpSurface;
	memset(lpDDMemory, 0, ddsd.dwHeight * ddsd.lPitch);
	lpBackSurface->Unlock(NULL);
}

void ZoomWindow(/*int zoomSet*/)
{
	TCHAR	szBuf[16];

//	zoom = zoomSet;
	if (!bFullScreen)
		curMode &= 1;
	curMode++;
	if (curMode >= numMode)
		curMode = 0;
	if (bFullScreen)
	{
		ReleaseSurfaces();
//		dwScreenWidth = 320*zoom;
//		dwScreenHeight = 240*zoom;
		InitSurfaces(bFullScreen);
		wsprintf(szBuf, _T("%dx%dx8"), mode[curMode].width, mode[curMode].height);
		SetInfoString(szBuf);
	}
	else
	{
		ReleaseSurfaces();
//		dwScreenWidth = 320*zoom;
//		dwScreenHeight = 240*zoom;
		int cx = 320*((curMode&1)+1) + GetSystemMetrics(SM_CXFIXEDFRAME)*2 + GetSystemMetrics(SM_CXEDGE)*2;
		int cy = 240*((curMode&1)+1) + GetSystemMetrics(SM_CYFIXEDFRAME)*2 + GetSystemMetrics(SM_CYEDGE)*2
					+ GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU);
		MoveWindow(hwnd, xPosWindowed, yPosWindowed, cx, cy, TRUE);
		ShowWindow(hwnd, SW_HIDE);
		ShowWindow(hwnd, SW_MINIMIZE);
		InitSurfaces(bFullScreen);
		ShowWindow(hwnd, SW_RESTORE);
		wsprintf(szBuf, _T("%dx%d"), 320*((curMode&1)+1), 240*((curMode&1)+1));
		SetInfoString(szBuf);
	}

	HRESULT	ddret;
	DDSURFACEDESC	ddsd;
	BYTE	*lpDDMemory;

	memset(&ddsd, 0, sizeof(DDSURFACEDESC));
	ddsd.dwSize = sizeof(ddsd);
	LogDumpDetail("DD:Lock(zoom)...");
	ddret = lpBackSurface->Lock(NULL, &ddsd, 0, NULL);
	LogDumpDetail("Done\n");
	if (FAILED(ddret))
	{
		LogDump("ZoomWindow - Locks [code=%08X]\n", ddret);
		return;
	}

	lpDDMemory = (BYTE *)ddsd.lpSurface;
	memset(lpDDMemory, 0, ddsd.dwHeight * ddsd.lPitch);
	lpBackSurface->Unlock(NULL);
}

void OnMoveWindow(int x, int y)
{
	GetWindowRect(hwnd, &rcWnd);
	rcWnd.left += GetSystemMetrics(SM_CXFIXEDFRAME) + GetSystemMetrics(SM_CXEDGE);
	rcWnd.top += GetSystemMetrics(SM_CXFIXEDFRAME) + GetSystemMetrics(SM_CXEDGE)
					+ GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU);
	rcWnd.right = rcWnd.left + 320*((curMode&1)+1);
	rcWnd.bottom = rcWnd.top + 240*((curMode&1)+1);
}
