// di.cpp

#include "stdafx.h"

#ifdef USE_JOYSTICK

#define DI_CPP
#include "di.h"

LPDIRECTINPUT			lpDInput = NULL;
LPDIRECTINPUTDEVICE2	lpJoystick = NULL;
GUID					JoystickGUID;

static BOOL CALLBACK DIEnumProc(LPCDIDEVICEINSTANCE lpddi, LPVOID pvRef);
static BOOL bDIDeviceFound = FALSE;

//#define DebugErrorMessage(mes)	{TCHAR szBuf[1024];wsprintf(szBuf, "%s (code=%d)\n", mes, diret);TRACE(szBuf);AfxMessageBox(szBuf);}
#define DebugErrorMessage		TRACE


BOOL InitDInput(void)
{
	HRESULT	diret;

	diret = DirectInputCreate(AfxGetApp()->m_hInstance, DIRECTINPUT_VERSION, &lpDInput, NULL);
	if (FAILED(diret))
	{
		lpDInput = NULL;
		DebugErrorMessage("Cannot create DirectInput\n");
		return FALSE;
	}

	bDIDeviceFound = FALSE;
	diret = lpDInput->EnumDevices(DIDEVTYPE_JOYSTICK, DIEnumProc, &JoystickGUID, DIEDFL_ATTACHEDONLY);
	if (FAILED(diret) || !bDIDeviceFound)
	{
		ReleaseDInput();
		DebugErrorMessage("Not found DirectInputDevice\n");
		return FALSE;
	}

	LPDIRECTINPUTDEVICE	lpJoyTemp;
	diret = lpDInput->CreateDevice(JoystickGUID, &lpJoyTemp, NULL);
	if (FAILED(diret))
	{
		ReleaseDInput();
		DebugErrorMessage("Cannot create DirectInputDevice\n");
		return FALSE;
	}

	diret = lpJoyTemp->QueryInterface(IID_IDirectInputDevice2, (LPVOID *)&lpJoystick);
	lpJoyTemp->Release();
	if (FAILED(diret))
	{
		ReleaseDInput();
		DebugErrorMessage("QueryInterface IID_IDirectInputDevice2 failed\n");
		return FALSE;
	}

	diret = lpJoystick->SetDataFormat(&c_dfDIJoystick);
	if (FAILED(diret))
	{
		ReleaseDevices();
		ReleaseDInput();
		DebugErrorMessage("SetDataFormat failed\n");
		return FALSE;
	}

	diret = lpJoystick->SetCooperativeLevel(AfxGetMainWnd()->GetSafeHwnd(),
							DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
	if (FAILED(diret))
	{
		ReleaseDevices();
		ReleaseDInput();
		DebugErrorMessage("SetCooperativeLevel failed\n");
		return FALSE;
	}

	DIPROPRANGE	diprg;

	diprg.diph.dwSize = sizeof(diprg);
	diprg.diph.dwHeaderSize = sizeof(diprg.diph);
	diprg.diph.dwObj = DIJOFS_X;
	diprg.diph.dwHow = DIPH_BYOFFSET;
	diprg.lMin = -AXIS_MAX;
	diprg.lMax = +AXIS_MAX;

	diret = lpJoystick->SetProperty(DIPROP_RANGE, &diprg.diph);
	if (FAILED(diret))
	{
		ReleaseDevices();
		ReleaseDInput();
		DebugErrorMessage("SetProperty failed (x range)\n");
		return FALSE;
	}

	diprg.diph.dwSize = sizeof(diprg);
	diprg.diph.dwHeaderSize = sizeof(diprg.diph);
	diprg.diph.dwObj = DIJOFS_Y;
	diprg.diph.dwHow = DIPH_BYOFFSET;
	diprg.lMin = -AXIS_MAX;
	diprg.lMax = +AXIS_MAX;

	diret = lpJoystick->SetProperty(DIPROP_RANGE, &diprg.diph);
	if (FAILED(diret))
	{
		ReleaseDevices();
		ReleaseDInput();
		DebugErrorMessage("SetProperty failed (y range)\n");
		return FALSE;
	}

	DIPROPDWORD	dipdw;

	dipdw.diph.dwSize = sizeof(dipdw);
	dipdw.diph.dwHeaderSize = sizeof(dipdw.diph);
	dipdw.diph.dwObj = DIJOFS_X;
	dipdw.diph.dwHow = DIPH_BYOFFSET;
	dipdw.dwData = 5000;

	diret = lpJoystick->SetProperty(DIPROP_DEADZONE, &dipdw.diph);
	if (FAILED(diret))
	{
		ReleaseDevices();
		ReleaseDInput();
		DebugErrorMessage("SetProperty failed (x deadzone)\n");
		return FALSE;
	}

	dipdw.diph.dwSize = sizeof(dipdw);
	dipdw.diph.dwHeaderSize = sizeof(dipdw.diph);
	dipdw.diph.dwObj = DIJOFS_Y;
	dipdw.diph.dwHow = DIPH_BYOFFSET;
	dipdw.dwData = 5000;

	diret = lpJoystick->SetProperty(DIPROP_DEADZONE, &dipdw.diph);
	if (FAILED(diret))
	{
		ReleaseDevices();
		ReleaseDInput();
		DebugErrorMessage("SetProperty failed (y deadzone)\n");
		return FALSE;
	}

	if (!AcquireInput())
		return FALSE;

	return TRUE;
}

BOOL CALLBACK DIEnumProc(LPCDIDEVICEINSTANCE lpddi, LPVOID pvRef)
{
	GUID	*lpGUID = (GUID *)pvRef;

	TRACE("Device found\n");

	*lpGUID = lpddi->guidInstance;
	bDIDeviceFound = TRUE;

	return DIENUM_STOP;
}

BOOL ReleaseDInput(void)
{
	if (lpDInput)
	{
		lpDInput->Release();
		lpDInput = NULL;
	}
	return TRUE;
}

BOOL ReleaseDevices(void)
{
	if (lpJoystick)
	{
		lpJoystick->Release();
		lpJoystick = NULL;
	}
	return TRUE;
}

BOOL AcquireInput(void)
{
	HRESULT	diret=0;

	if (!lpJoystick)
	{
		DebugErrorMessage("AcquireInput: lpJoystick = NULL\n");
		return FALSE;
	}

	diret = lpJoystick->Acquire();
	if (FAILED(diret))
	{
		DebugErrorMessage("AcquireInput: Acquire failed\n");
		return FALSE;
	}
	return TRUE;
}

#endif // USE_JOYSTICK
