// ds.cpp

#include "stdafx.h"
#ifdef SOUND
#include <process.h>
#define DS_CPP
#include "ds.h"
#include "log.h"

LPDIRECTSOUND		lpDSound		= NULL;
LPDIRECTSOUNDBUFFER	lpSoundBuffer	= NULL;
LPDIRECTSOUNDNOTIFY	lpSoundNotify	= NULL;
LPDIRECTSOUNDBUFFER	lpPrimarySoundBuffer	= NULL;
int		ds_nBits		= 16;
int		ds_nChannels	= 2;
int		ds_nSampleRate	= 44100;
int		ds_nLockLatency	= 40;
DWORD	ds_dwLockSize;
DWORD	ds_dwBufferSize;
int		ds_nPositions	= 0;
LPDSBPOSITIONNOTIFY	ds_pdsbpn	= NULL;
BOOL	ds_bUsePolling	= FALSE;
WORD	**ds_buf = NULL;
size_t	ds_bufsize = 0;
int		ds_CoopMode		= 2;

extern BOOL SupportMMX;

static HANDLE	ds_hThread = NULL;
static DWORD	ds_dwThreadId = NULL;
static DWORD	ds_dwWritePos = 0;

static HANDLE	hEventStop = NULL;
static HANDLE	hEventWriteBuffer = NULL;

static UINT CALLBACK PlayProc(LPVOID lpParam);
static BOOL WriteBuffer(void);
static void AddBuffer(WORD *buf1, const WORD *buf2, DWORD dwSize);
extern void WriteSoundData(WORD *buf, int ch, DWORD dwSize);

BOOL InitDSound(HWND hwnd)
{
	HRESULT	dsret;
	int		i;
	WAVEFORMATEX	wfx;

	LogDumpDetail("DS:DirectSoundCreate...");
	dsret = DirectSoundCreate(NULL, &lpDSound, NULL);
	LogDumpDetail("Done\n");
	if (FAILED(dsret))
	{
		LogDump("InitDSound - Cannot create DirectSound [code=%08X]\n", dsret);
		return FALSE;
	}

	LogDumpDetail("DS:SetCooperativeLevel...");
	switch (ds_CoopMode)
	{
	case 0:
		dsret = lpDSound->SetCooperativeLevel(hwnd, DSSCL_WRITEPRIMARY);
		break;
	case 1:
		dsret = lpDSound->SetCooperativeLevel(hwnd, DSSCL_EXCLUSIVE);
		break;
	default:
		dsret = lpDSound->SetCooperativeLevel(hwnd, DSSCL_PRIORITY);
		break;
	}
	LogDumpDetail("Done\n");
	if (FAILED(dsret))
	{
		LogDump("InitDSound - SetCooperativeLevel failed [code=%08X]\n", dsret);
		ReleaseDSound();
		return FALSE;
	}

	DSBUFFERDESC	dsbd;

	if (ds_CoopMode != 0)
	{
		ZeroMemory(&dsbd, sizeof(DSBUFFERDESC));
		dsbd.dwSize = sizeof(DSBUFFERDESC);
		dsbd.dwBufferBytes = 0;
		dsbd.lpwfxFormat = NULL;
		dsbd.dwFlags = /*DSBCAPS_CTRLVOLUME | DSBCAPS_STICKYFOCUS |*/ DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_PRIMARYBUFFER;
		LogDumpDetail("DS:CreateSoundBuffer(primary)...");
		dsret = lpDSound->CreateSoundBuffer(&dsbd, &lpPrimarySoundBuffer, NULL);
		LogDumpDetail("Done\n");
		if (FAILED(dsret))
		{
			LogDump("InitDSound - CreateSoundBuffer(primary) failed [code=%08X]\n", dsret);
			ReleaseDSound();
			return FALSE;
		}

		DSBCAPS	dsbc;
		ZeroMemory(&dsbc, sizeof(DSBCAPS));
		dsbc.dwSize = sizeof(DSBCAPS);
		LogDumpDetail("DS:GetCaps...");
		dsret = lpPrimarySoundBuffer->GetCaps(&dsbc);
		LogDumpDetail("Done\n");
		if (FAILED(dsret))
		{
			LogDump("InitDSound - IDirectSoundBuffer::GetCaps failed [code=%08X]\n",dsret);
			ReleaseDSound();
			return FALSE;
		}
		ds_dwBufferSize = dsbc.dwBufferBytes;

		wfx.wFormatTag = WAVE_FORMAT_PCM;
		wfx.nChannels = ds_nChannels;
		wfx.nSamplesPerSec = ds_nSampleRate;
		wfx.nAvgBytesPerSec = ds_nBits/8*ds_nChannels*ds_nSampleRate;;
		wfx.nBlockAlign = ds_nBits/8*ds_nChannels;
		wfx.wBitsPerSample = ds_nBits;
		wfx.cbSize = sizeof(WAVEFORMATEX);
	}

	ZeroMemory(&dsbd, sizeof(DSBUFFERDESC));
	dsbd.dwSize = sizeof(DSBUFFERDESC);
	if (ds_CoopMode == 0)
	{
		dsbd.dwBufferBytes = 0;
		dsbd.lpwfxFormat = NULL;
		dsbd.dwFlags = /*DSBCAPS_CTRLVOLUME |*/ DSBCAPS_STICKYFOCUS | DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_PRIMARYBUFFER;
	}
	else
	{
		dsbd.dwBufferBytes = ds_dwBufferSize;
		dsbd.lpwfxFormat = &wfx;
		dsbd.dwFlags = /*DSBCAPS_CTRLVOLUME |*/ DSBCAPS_STICKYFOCUS | DSBCAPS_GETCURRENTPOSITION2;;
	}
	LogDumpDetail("DS:CreateSoundBuffer...");
	dsret = lpDSound->CreateSoundBuffer(&dsbd, &lpSoundBuffer, NULL);
	LogDumpDetail("Done\n");
	if (FAILED(dsret))
	{
		LogDump("InitDSound - CreateSoundBuffer failed [code=%08X]\n", dsret);
		ReleaseDSound();
		return FALSE;
	}

	if (ds_CoopMode == 0)
	{
		DSBCAPS	dsbc;
		ZeroMemory(&dsbc, sizeof(DSBCAPS));
		dsbc.dwSize = sizeof(DSBCAPS);
		LogDumpDetail("DS:GetCaps(writeprimary)...");
		dsret = lpSoundBuffer->GetCaps(&dsbc);
		LogDumpDetail("Done\n");
		if (FAILED(dsret))
		{
			LogDump("InitDSound - IDirectSoundBuffer::GetCaps failed (writeprimary) [code=%08X]\n",dsret);
			ReleaseDSound();
			return FALSE;
		}
		ds_dwBufferSize = dsbc.dwBufferBytes;
	}

//	dsret = lpSoundBuffer->QueryInterface(IID_IDirectSoundNotify, (void **)&lpSoundNotify);
//	if (FAILED(dsret))
	{
		TRACE("InitDSound - Cannnot get IDirectSoundNotify(code=%X)\n", dsret);
		ds_bUsePolling = TRUE;
		lpSoundNotify = NULL;
	}

	DSCAPS	dsc;
	ZeroMemory(&dsc, sizeof(DSCAPS));
	dsc.dwSize = sizeof(DSCAPS);
	LogDumpDetail("DS:GetCaps(dsound)...");
	dsret = lpDSound->GetCaps(&dsc);
	LogDumpDetail("Done\n");
	if (FAILED(dsret))
	{
		LogDump("InitDSound - IDirectSound::GetCaps failed [code=%08X]\n", dsret);
		ReleaseDSound();
		return FALSE;
	}
//	if (!(dsc.dwFlags&DSCAPS_PRIMARY16BIT) && ds_nBits == 16)
//		ds_nBits = 8;
	if (!(dsc.dwFlags&DSCAPS_PRIMARYSTEREO) && ds_nChannels == 2)
		ds_nChannels = 1;

	wfx.wFormatTag = WAVE_FORMAT_PCM;
	wfx.nChannels = ds_nChannels;
	wfx.nSamplesPerSec = ds_nSampleRate;
	wfx.nAvgBytesPerSec = ds_nBits/8*ds_nChannels*ds_nSampleRate;;
	wfx.nBlockAlign = ds_nBits/8*ds_nChannels;
	wfx.wBitsPerSample = ds_nBits;
	wfx.cbSize = sizeof(WAVEFORMATEX);
	LogDumpDetail("DS:SetFormat...");
	if (lpPrimarySoundBuffer != NULL)
	{
		DWORD	dwWritten;

		dsret = lpPrimarySoundBuffer->SetFormat(&wfx);
		lpPrimarySoundBuffer->GetFormat(&wfx, sizeof(wfx), &dwWritten);
	}
	else
	{
		DWORD	dwWritten;

		dsret = lpSoundBuffer->SetFormat(&wfx);
		lpSoundBuffer->GetFormat(&wfx, sizeof(wfx), &dwWritten);
	}
	LogDumpDetail("Done\n");
	if (FAILED(dsret))
	{
		LogDump("InitDSound - SetFormat failed [code=%08X]\n", dsret);
		ReleaseDSound();
		return FALSE;
	}
	if (wfx.nSamplesPerSec != (DWORD)ds_nSampleRate)
	{
		LogDump("Unsupported sample rate\n");
		ReleaseDSound();
		return FALSE;
	}

	if (lpPrimarySoundBuffer != NULL)
	{
		LogDumpDetail("DS:SetFormat(primary)...");
		dsret = lpPrimarySoundBuffer->SetFormat(&wfx);
		LogDumpDetail("Done\n");
		if (FAILED(dsret))
		{
			LogDump("InitDSound - SetFormat(primary) failed [code=%08X]\n", dsret);
			ReleaseDSound();
			return FALSE;
		}
	}

	int	nBytesPerSec = ds_nBits/8*ds_nChannels*ds_nSampleRate;
	ds_dwLockSize = nBytesPerSec*ds_nLockLatency/1000;
	TRACE("ds_dwLockSize = %X\n", ds_dwLockSize);
	ds_dwLockSize = min(ds_dwLockSize, ds_dwBufferSize/2);
	ds_dwLockSize = ds_dwLockSize - ds_dwLockSize%(ds_nBits/8*ds_nChannels);

	if (!ds_bUsePolling)
	{
		ds_nPositions = (ds_dwBufferSize+ds_dwLockSize-1) / ds_dwLockSize;		// ؂グ
		ds_dwLockSize = (ds_dwBufferSize+ds_nPositions-1) / ds_nPositions;

		ds_pdsbpn = (LPDSBPOSITIONNOTIFY)calloc(ds_nPositions, sizeof(DSBPOSITIONNOTIFY));
		if (ds_pdsbpn == NULL)
		{
			LogDump("InitDSound - Out of memory in calloc DSBPOSITIONNOTIFY\n");
			ds_nPositions = 0;
			ReleaseDSound();
			return FALSE;
		}
		for (i = 0; i < ds_nPositions; i++)
		{
			ds_pdsbpn[i].dwOffset = i * ds_dwLockSize;
			ds_pdsbpn[i].hEventNotify = CreateEvent(NULL, FALSE, FALSE, NULL);
			if (ds_pdsbpn[i].hEventNotify == NULL)
			{
				LogDump("InitDSound - CreateEvent(hEventNotify[%d]) failed\n", i);
				ReleaseDSound();
				return FALSE;
			}
		}
		LogDumpDetail("DS:SetNortificationPositions...");
		dsret = lpSoundNotify->SetNotificationPositions(ds_nPositions, ds_pdsbpn);
		LogDumpDetail("Done\n");
		if (FAILED(dsret))
		{
			LogDump("InitDSound - IDirectSoundNotify::SetNotificationPositions failed [code=%08X]\n", dsret);
			ReleaseDSound();
			return FALSE;
		}
	}

	hEventStop = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (hEventStop == NULL)
	{
		LogDump("InitDSound - CreateEvent(hEventStop) failed\n");
		ReleaseDSound();
		return FALSE;
	}
	hEventWriteBuffer = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (hEventWriteBuffer == NULL)
	{
		LogDump("InitDSound - CreateEvent(hEventWriteBuffer) failed\n");
		ReleaseDSound();
		return FALSE;
	}

	DWORD	dwPlayCursor;
	LogDumpDetail("DS:GetCurrentPosition...");
	dsret = lpSoundBuffer->GetCurrentPosition(&dwPlayCursor, &ds_dwWritePos);
	LogDumpDetail("Done\n");
	if (FAILED(dsret))
	{
		LogDump("InitDSound - GetCurrentPosition failed [code=%08X]\n", dsret);
		ReleaseDSound();
		return FALSE;
	}

	if (!StartSound())
	{
		ReleaseDSound();
		return FALSE;
	}

	return TRUE;
}

BOOL StartSound(void)
{
	HRESULT	dsret;

	ResetEvent(hEventStop);
	ResetEvent(hEventWriteBuffer);
	ds_hThread = (HANDLE)_beginthreadex(NULL, 0, PlayProc, NULL, 0, (UINT *)&ds_dwThreadId);
	if (ds_hThread == NULL)
	{
		LogDump("StartSound - _beginthreadex failed\n");
		return FALSE;
	}
	if (!SetThreadPriority(ds_hThread, THREAD_PRIORITY_ABOVE_NORMAL))
	{
		LogDump("StartSound - SetThreadPriority faild\n");
		return FALSE;
	}

	if (!WriteBuffer())
		return FALSE;

	if (lpPrimarySoundBuffer != NULL)
	{
		LogDumpDetail("DS:Play(primary)...");
		dsret = lpPrimarySoundBuffer->Play(0, 0, DSBPLAY_LOOPING);
		LogDumpDetail("Done\n");
		if (FAILED(dsret))
		{
			LogDump("StartSound - Play(primary) failed [code=%08X]\n", dsret);
			ReleaseDSound();
			return FALSE;
		}
	}

	LogDumpDetail("DS:Play...");
	dsret = lpSoundBuffer->Play(0, 0, DSBPLAY_LOOPING);
	LogDumpDetail("Done\n");
	if (FAILED(dsret))
	{
		LogDump("StartSound - Play failed [code=%08X]\n", dsret);
		ReleaseDSound();
		return FALSE;
	}

	SetEvent(hEventWriteBuffer);

	return TRUE;
}

UINT CALLBACK PlayProc(LPVOID lpParam)
{
	HRESULT	dsret;
	int		i;
	DWORD	dwObject;
	HANDLE	aHandles[] = {hEventWriteBuffer, hEventStop};

	dwObject = WaitForMultipleObjects(2, aHandles, FALSE, INFINITE);
	if (dwObject != WAIT_OBJECT_0)
		return 0;

	if (ds_bUsePolling)
	{
		DWORD	dwPlayCursor;
		DWORD	dwWriteCursor;
		DWORD	dwLength;

		while (TRUE)
		{
			dwObject = WaitForSingleObject(hEventStop, ds_nLockLatency/2);
			if (dwObject == WAIT_OBJECT_0)
				break;
			if (lpSoundBuffer != NULL)
			{
				do
				{
					dsret = lpSoundBuffer->GetCurrentPosition(&dwPlayCursor, &dwWriteCursor);
					if (FAILED(dsret))
					{
						LogDump("PlayProc - GetCurrentPosition failed [code=%08X]\n", dsret);
						return 0;
					}
					dwLength = ds_dwWritePos - dwPlayCursor;
					if ((int)dwLength < 0)
					{
						//TRACE("l=%X\n", dwLength);
						dwLength += ds_dwBufferSize;
					}
					//TRACE("dwLength = %X\n", dwLength);
					//TRACE("p=%X,w=%X\n", dwPlayCursor, dwWriteCursor);
					//TRACE("dwLength = %X\n", dwLength);
					if (dwLength < ds_dwLockSize)
						WriteBuffer();
				} while (dwLength < ds_dwLockSize);
			}
		}
	}
	else
	{
		HANDLE	*aHandlesNotify;
		int		iPos;

		aHandlesNotify = (HANDLE *)malloc((ds_nPositions+1) * sizeof(HANDLE));
		if (aHandlesNotify == NULL)
		{
			LogDump("PlayProc - Out of memory in malloc HANDLES\n");
			return 1;
		}
		for (i = 0; i < ds_nPositions; i++)
			aHandlesNotify[i] = ds_pdsbpn[i].hEventNotify;
		aHandlesNotify[ds_nPositions] = hEventStop;
		while (TRUE)
		{
			dwObject = WaitForMultipleObjects(ds_nPositions + 1, aHandlesNotify, FALSE, INFINITE);
			iPos = dwObject - WAIT_OBJECT_0;
			if (iPos == ds_nPositions)
				break;
			if (iPos < 0 || ds_nPositions <= iPos)
				continue;
			if (lpSoundBuffer != NULL)
			{
				ds_dwWritePos = ds_pdsbpn[(iPos+1)%ds_nPositions].dwOffset;
				WriteBuffer();
			}
		}
		free(aHandlesNotify);
	}

	return 0;
}

BOOL WriteBuffer(void)
{
	HRESULT	dsret;
	int		i;

	WORD	*lpAudio1;
	WORD	*lpAudio2;
	DWORD	dwAudioBytes1;
	DWORD	dwAudioBytes2;

/*	TRACE("lpAudio1 = %X\n"
		  "dwAudioBytes1 = %X\n"
		  "lpAudio2 = %X\n"
		  "dwAudioBytes2 = %X\n", lpAudio1, dwAudioBytes1, lpAudio2, dwAudioBytes2);
	TRACE("ds_dwLockSize = %X\n", ds_dwLockSize);
*/
	DWORD	dwAudioBytes;

	dsret = lpSoundBuffer->Lock(ds_dwWritePos, ds_dwLockSize, (void **)&lpAudio1, &dwAudioBytes1,
								(void **)&lpAudio2, &dwAudioBytes2, 0);
	if (dsret == DSERR_BUFFERLOST)
	{
		dsret = lpSoundBuffer->Restore();
		if (FAILED(dsret))
		{
			LogDump("WriteBuffer - Restore failed [code=%08X]\n", dsret);
			return FALSE;
		}
		dsret = lpSoundBuffer->Lock(ds_dwWritePos, ds_dwLockSize, (void **)&lpAudio1, &dwAudioBytes1,
									(void **)&lpAudio2, &dwAudioBytes2, 0);
	}
	if (FAILED(dsret))
	{
		LogDump("WriteBuffer - Lock failed [code=%08X]\n", dsret);
		return FALSE;
	}

	dwAudioBytes = dwAudioBytes1 + (lpAudio2 ? dwAudioBytes2 : 0);
	if (ds_bufsize < dwAudioBytes)
	{
		WORD **tmp_ptr = (WORD **)realloc(ds_buf, 6*sizeof(char *) + 6*dwAudioBytes);
		if (tmp_ptr == NULL)
			goto WRITE_BUFFER_ERROR_RETURN;
		ds_buf = tmp_ptr;
		for (i = 0; i < 6; i++)
			ds_buf[i] = (WORD *)((char *)tmp_ptr + 6*sizeof(char *) + i*dwAudioBytes);
		ds_bufsize = dwAudioBytes;
	}
	for (i = 0; i < 6; i++)
		WriteSoundData(ds_buf[i], i, dwAudioBytes/2);
	for (i = 1; i < 6; i++)
		AddBuffer(ds_buf[0], ds_buf[i], dwAudioBytes/2);

	DWORD	dwPos;
	DWORD	dwPos2;

	for (dwPos = 0; dwPos < dwAudioBytes1/2; dwPos++)
		lpAudio1[dwPos] = ds_buf[0][dwPos];
	if (lpAudio2 != NULL)
	{
		for (dwPos2 = 0; dwPos2 < dwAudioBytes2/2; dwPos2++, dwPos++)
			lpAudio2[dwPos2] = ds_buf[0][dwPos];
	}

	lpSoundBuffer->Unlock(lpAudio1, dwAudioBytes1, lpAudio2, dwAudioBytes2);

WRITE_BUFFER_ERROR_RETURN:
	ds_dwWritePos += dwAudioBytes;
	if (ds_dwWritePos >= ds_dwBufferSize)
		ds_dwWritePos -= ds_dwBufferSize;

	return TRUE;
}

void AddBuffer(WORD *buf1, const WORD *buf2, DWORD dwSize)
{
	if (!SupportMMX)
	{
		DWORD	i;

		for (i = 0; i < dwSize; i++)
			*buf1++ += *buf2++;
	}
	else
	{
		__asm {
			mov		ebx,buf2
			mov		eax,buf1
			mov		edx,8
			mov		ecx,dwSize
			shr		ecx,2
		lp1:
			movq	mm0,[eax]
			paddsw	mm0,[ebx]
			movq	[eax],mm0
			add		ebx,edx
			add		eax,edx
			loop	lp1
			mov		ecx,dwSize
			and		ecx,3
			je		nx1
		lp2:
			mov		dx,[eax]
			add		dx,[ebx]
			mov		[eax],dx
			add		ebx,2
			add		eax,2
			loop	lp2
		nx1:
			emms
		}
	}
}

BOOL ReleaseDSound(void)
{
	if (lpDSound != NULL)
	{
		ReleaseSoundBuffer();
		LogDumpDetail("DS:Release(dsound)...");
		lpDSound->Release();
		LogDumpDetail("Done\n");
		lpDSound = NULL;
	}
	if (hEventStop != NULL)
	{
		CloseHandle(hEventStop);
		hEventStop = NULL;
	}
	if (hEventWriteBuffer != NULL)
	{
		CloseHandle(hEventWriteBuffer);
		hEventWriteBuffer = NULL;
	}
	if (ds_buf != NULL)
	{
		free(ds_buf);
		ds_bufsize = 0;
	}
	return TRUE;
}

BOOL ReleaseSoundBuffer(void)
{
	int		i;

	if (lpSoundNotify != NULL)
	{
		LogDumpDetail("DS:Release(notify)...");
		lpSoundNotify->Release();
		LogDumpDetail("Done\n");
	}
	if (lpSoundBuffer != NULL)
	{
		StopSound();
		LogDumpDetail("DS:Release(soundbuffer)...");
		lpSoundBuffer->Release();
		LogDumpDetail("Done\n");
		lpSoundBuffer = NULL;
	}
	if (ds_nPositions > 0)
	{
		for (i = 0; i < ds_nPositions; i++)
		{
			if (ds_pdsbpn[i].hEventNotify != NULL)
			{
				CloseHandle(ds_pdsbpn[i].hEventNotify);
				ds_pdsbpn[i].hEventNotify = NULL;
			}
		}
		free(ds_pdsbpn);
		ds_pdsbpn = NULL;
		ds_nPositions = 0;
	}
	return TRUE;
}

BOOL StopSound(void)
{
	HRESULT	dsret;
	DWORD	dwStatus;

	if (lpSoundBuffer != NULL)
	{
		SetEvent(hEventStop);
		dsret = lpSoundBuffer->GetStatus(&dwStatus);
		if (FAILED(dsret))
		{
			LogDump("StopSound - GetStatus failed [code=%08X]\n", dsret);
			return FALSE;
		}
		if (dwStatus & DSBSTATUS_PLAYING)
		{
			LogDumpDetail("DS:Stop...");
			dsret = lpSoundBuffer->Stop();
			LogDumpDetail("Done\n");
			if (FAILED(dsret))
			{
				LogDump("StopSound - Stop failed [code=%08X]\n", dsret);
				return FALSE;
			}

			if (lpPrimarySoundBuffer != NULL)
			{
				LogDumpDetail("DS:Stop(primary)...");
				dsret = lpPrimarySoundBuffer->Stop();
				LogDumpDetail("Done\n");
				if (FAILED(dsret))
				{
					LogDump("StopSound - Stop(primary) failed [code=%08X]\n", dsret);
					return FALSE;
				}
			}
		}
		if (ds_hThread != NULL)
		{
			LogDumpDetail("DS:waiting thread termination...");
			WaitForSingleObject(ds_hThread, INFINITE);
			LogDumpDetail("Done\n");
			CloseHandle(ds_hThread);
			ds_hThread = NULL;
		}
	}
	return TRUE;
}

#endif // SOUND
